#include <stdbool.h>
#include <string.h>

#include "ltm/common.h"
#include "ltm/template.h"
#include "ltm/template_internal.h"

ltm_err ltm_instance_block_init(ltm_instance_block **out) {
  ltm_instance_block *block = calloc(1, sizeof(ltm_instance_block));

  if (block == NULL) {
    return ltm_err_failed_alloc;
  }

  *out = block;

  return ltm_err_ok;
}

ltm_err ltm_instance_init(ltm_instance **out) {
  ltm_instance *instance = calloc(1, sizeof(ltm_instance));

  if (instance == NULL) {
    return ltm_err_failed_alloc;
  }

  *out = instance;

  return ltm_err_ok;
}

void ltm_instance_free(ltm_instance *instance) {
  if (instance->vars != NULL) {
    free(instance->vars);
  }

  ltm_instance_block *block = instance->blocks.head;

  while (block != NULL) {
    switch (block->type) {
    case ltm_instance_block_type_buf_owned:
      free(block->data.ptr);
      break;
    case ltm_instance_block_type_nested:
      ltm_instance_free(block->data.ptr);
      break;
    case ltm_instance_block_type_buf:;
    }

    ltm_instance_block *temp = block->next;
    free(block);
    block = temp;
  }

  free(instance);
}

ltm_err ltm_template_instantiate(ltm_instance **out,
                                 const ltm_template *template) {
  ltm_instance *instance;
  LTM_RES(ltm_instance_init(&instance));

  instance->template = template;

  ltm_instance_block ***vars = NULL;

  if (template->names.len > 0) {
    vars = malloc(template->names.len * sizeof(ltm_instance_block **));

    if (vars == NULL) {
      ltm_instance_free(instance);

      return ltm_err_failed_alloc;
    }

    instance->vars = vars;
  }

  for (size_t block_index = 0; block_index < template->blocks.len;
       block_index++) {
    ltm_template_block *template_block = &template->blocks.arr[block_index];

    switch (template_block->type) {
    case ltm_template_block_type_literal: {
      ltm_instance_block *block;
      LTM_RES2(ltm_instance_block_init(&block), ltm_instance_free(instance));

      block->type = ltm_instance_block_type_buf;
      block->data.ptr = template_block->data.ptr;
      block->data.len = template_block->data.len;

      if (instance->blocks.head == NULL) {
        instance->blocks.head = block;
        instance->blocks.tail = block;
        instance->blocks.current = block;
      } else {
        instance->blocks.tail->next = block;
        instance->blocks.tail = block;
      }
    } break;
    case ltm_template_block_type_var:
    case ltm_template_block_type_nested:
      *vars = &instance->blocks.tail->next;
      vars++;
      break;
    }
  }

  *out = instance;

  return ltm_err_ok;
}

ltm_err ltm_instance_block_add_var(ltm_instance *instance, const char *name,
                                   ltm_instance_block_type type, void *data,
                                   size_t len) {
  const ltm_template *template = instance->template;
  bool matched = false;

  size_t i = 0;
  for (i = 0; i < template->names.len && !matched; i++) {
    ltm_template_block_name *block_name = &template->names.arr[i];
    matched = strncmp(name, block_name->name.s, block_name->name.len) == 0;
  }

  if (!matched) {
    return ltm_err_not_found;
  }

  ltm_template_block *template_block = &template->blocks.arr[i];

  if (template_block->type != ltm_template_block_type_var) {
    return ltm_err_wrong_block_type;
  }

  ltm_instance_block *block;
  LTM_RES(ltm_instance_block_init(&block));

  block->type = type;
  block->data.ptr = data;
  block->data.len = len;

  // We insert the block in the linked list and replace its next pointer as the
  // new attachment point for this variable
  block->next = *instance->vars[i];
  *instance->vars[i] = block;
  instance->vars[i] = &block->next;

  return ltm_err_ok;
}

ltm_err ltm_instance_block_add_nested(ltm_instance **out,
                                      ltm_instance *instance,
                                      const char *name) {
  const ltm_template *template = instance->template;
  bool matched = false;

  size_t i = 0;
  for (i = 0; i < template->names.len && !matched; i++) {
    ltm_template_block_name *block_name = &template->names.arr[i];
    matched = strncmp(name, block_name->name.s, block_name->name.len) == 0;
  }

  if (!matched) {
    return ltm_err_not_found;
  }

  ltm_template_block *template_block = &template->blocks.arr[i];

  if (template_block->type != ltm_template_block_type_nested) {
    return ltm_err_wrong_block_type;
  }

  ltm_instance_block *block;
  LTM_RES(ltm_instance_block_init(&block));

  ltm_instance *nested;
  LTM_RES2(ltm_template_instantiate(&nested, template_block->data.ptr),
           free(block));

  block->type = ltm_instance_block_type_nested;
  block->data.ptr = nested;

  // We insert the block in the linked list and replace its next pointer as the
  // new attachment point for this variable
  block->next = *instance->vars[i];
  *instance->vars[i] = block;
  instance->vars[i] = &block->next;

  *out = nested;

  return ltm_err_ok;
}
